/*
 * usbhw.c : LPC2148 USB logic driver
 *   by uratan! 2008.9.22
 */
#include "common.h"
#include "LPC214x.h"
#include "pcon.h"
#include "irq.h"
#include "swi.h"
#include "dscr.h.S"
#include "usbreg.h"
#include "usbhw.h"

/*
 * my local debug flag
 */
#define UDEBUG 2
#define RPRINTF  if(UDEBUG>=2)rprintf		/* 2: more verbose */
#define rPRINTF  if(UDEBUG>=1)rprintf		/* 1: error report only */

/*
 * functions
 */
int uhw_Init(void);
void uhw_Connect(int con);
void uhw_WakeUp(void);
void uhw_SetAddress(u_char adr);
void uhw_Configure(int cfg);
void uhw_ConfigEP(u_char *pEPD);
void uhw_EnableEP(u_char EPNum);
void uhw_DisableEP(u_char EPNum);
void uhw_ResetEP(u_char EPNum);
void uhw_SetStallEP(u_char EPNum);
void uhw_ClrStallEP(u_char EPNum);
void uhw_WriteEP(u_char EPNum, u_char *pData, u_long cnt);
void uhw_arm_WriteEP(u_char EPNum);
u_long uhw_ReadEP(u_char EPNum, u_char *pData, int buf_size);
void uhw_ReadEP_finish(u_char EPNum);
static void ucmd_write_data(u_long cmd, u_long val);
static u_long ucmd_read_data(u_long cmd);
static void ucmd_no_data(u_long cmd);
static u_long get_phy_EPaddr(u_long EPNum);
static void USB_ISR(void) __attribute__ ((interrupt("IRQ")));

/*
 * externs
 */
/* in pdown.c */
extern void pdown_usb_pll_init(void);
/* in usbcore.c */
extern void ucore_ResetCore(void);
extern void ucore_isr_EP0_in(u_long unused);
extern void ucore_isr_EP0_out(u_long unused);
/* in periph.c */
extern const INTR_subFUNCs USB_Jump_Table[N_JT];
extern const INTR_subFUNCs USB_Jump_Table_EP[LPC_N_EP];

/*
 * variables
 */
/* in main.c */
extern int GotSUD;		/* SetUp Data */
extern int GotCOD;		/* Control-Out Data */
extern int Sleep;		/* shall sleep becaue in Suspend mode */


/* =============================================================== */

/*
 * USB Initialize Function
 *   Return Value:    None
 */
int uhw_Init(void)
{
	/*
	 * Turn on USB logic
	 */
	PCONP |= PP_PUSB;			/* Turn On USB PCLK */

	/*
	 * Enable Vbus sense (P0.23)
	 */
	PINSEL1 |= (1<<14);

	/*
	 * select P0.31 pin function
	 *   1 - GoodLinc
	 *   2 - ~Connect
	 */
	PINSEL1 |= (2<<30);

	/*
	 * Configure 48MHz USB Clock
	 */
	pdown_usb_pll_init();

	/*
	 * setup vectored interrupt
	 */
	if(install_irq(USB_INT, (void *)USB_ISR) == FALSE) {
		return(FALSE);
	}

	/*
	 * setup hardware buffer for Control Endpoint
	 */
	DEV_INT_CLR = UDI_EP_RLZED_INT;

	REALIZE_EP = 0x03;		/* reset value, not needed */
	EP_INDEX = 0;
	MAXPACKET_SIZE = DSCR_MPSIZE_EP0;
	EP_INDEX = 1;
	MAXPACKET_SIZE = DSCR_MPSIZE_EP0;

	while((DEV_INT_STAT & UDI_EP_RLZED_INT) == 0)
		;
		/*
		 * MAX_PACKET_SIZEs of each endpoint are not reset by USB-RESET,
		 * it seems that only REALIZE_EP register may be reset by USB-RESET.
		 */

	/*
	 * Enable USB interrupt
	 */
	EP_INT_PRIO = 0x00000000;
	EP_INT_CLR  = 0xFFFFFFFF;
	EP_INT_EN   = 0xFFFFFFFF;		/* <== see note1 */
	DEV_INT_PRIO = 0x00000000;
	DEV_INT_CLR  = 0xFFFFFFFF;
	DEV_INT_EN   = UDI_DEV_STAT_INT |
					((USB_Jump_Table[USB_JT_SOF] != NULL) ? UDI_FRAME_INT : 0) |
					((USB_Jump_Table[USB_JT_ERROR] != NULL) ? UDI_ERR_INT : 0) |
					UDI_EP_SLOW_INT;
	USB_INT_STAT |= (1<<31);
		/*
		 * note1:
		 *   EP_INT_EN register will NOT mask interrupt requests !!
		 *   only switch 'slave-transfer-mode' and 'normal-mode'
		 * note2:
		 *   EP_INT_CLR will generate reponse on CMD_DATA register !
		 * note3:
		 *   to clear UDI_EP_SLOW_INT, the condition 'EP_INT_STAT==0' needed.
		 *   and UDI_EP_SLOW_INT is NOT cleared by clearing EP_INT_STAT.
		 *     (you shall clear EP_INT_STAT first, then UDI_EP_SLOW_INT)
		 */

	return(TRUE);
}

/*
 * USB Connect Function
 *   Parameters:      con:   Connect/Disconnect
 *   Return Value:    None
 */
void uhw_Connect(int con)
{
	u_long imf;

	imf = IntDisable();

	ucmd_write_data(CMD_SET_DEV_STAT, DAT_WR_BYTE(con ? DSTAT_CONN : 0));

	IntRestore(imf);

	return;
}

/*
 * USB Remote Wakeup Function
 *   Return Value:    None
 */
void uhw_WakeUp(void)
{
	u_long imf;

	imf = IntDisable();

	ucmd_write_data(CMD_SET_DEV_STAT, DAT_WR_BYTE(0 | DSTAT_CONN));
		/*
		 * I want to write 0 to DSTAT_SUSP bit to do 'remote-wakeup' operation.
		 * if DSTAT_SUSP bit is already 0, this is ignored.
		 */

	IntRestore(imf);

	return;
}

/*
 * USB Set Address Function
 *   Parameters:      adr:   USB Address
 *   Return Value:    None
 */
void uhw_SetAddress(u_char adr)
{
	u_long imf;

	imf = IntDisable();

	ucmd_write_data(CMD_SET_ADDR, DAT_WR_BYTE(0x80 | adr)); /* Don't wait for next */
	ucmd_write_data(CMD_SET_ADDR, DAT_WR_BYTE(0x80 | adr)); /* Setup Status Phase */

	IntRestore(imf);

	return;
}

/*
 * USB Configure Function
 *   Parameters:      cfg:   Configure/Deconfigure
 *   Return Value:    None
 */
void uhw_Configure(int cfg)
{
	u_long imf;

	imf = IntDisable();

	ucmd_write_data(CMD_CFG_DEV, DAT_WR_BYTE(cfg ? 1 : 0));

	DEV_INT_CLR = UDI_EP_RLZED_INT;
	REALIZE_EP = 0x00000003;
	while((DEV_INT_STAT & UDI_EP_RLZED_INT) == 0)
		;

	IntRestore(imf);

	return;
}

/*
 * Configure USB Endpoint
 *   Parameters:      pEPD:  Pointer to Endpoint Descriptor
 *   Return Value:    None
 */
void uhw_ConfigEP(u_char *pEPD)
{
	u_long num, packet_size;
	u_long imf;

	num = get_phy_EPaddr(pEPD[/*bEndpointAddress*/ 2]);
	packet_size = (pEPD[4]<<0) + (pEPD[5]<<8);		/* wMaxPacketSize */

	imf = IntDisable();

	DEV_INT_CLR = UDI_EP_RLZED_INT;
	REALIZE_EP |= (1 << num);
	EP_INDEX = num;
	MAXPACKET_SIZE = packet_size;
	while((DEV_INT_STAT & UDI_EP_RLZED_INT) == 0)
		;

	IntRestore(imf);

	return;
}

/*
 * Enable USB Endpoint
 *   Parameters:      EPNum: Endpoint Number
 *                    EPNum.0..3: Address
 *                    EPNum.7:    Dir
 *   Return Value:    None
 */
void uhw_EnableEP(u_char EPNum)
{
	u_long imf;

	imf = IntDisable();

	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(0));

	IntRestore(imf);

	return;
}

/*
 * Disable USB Endpoint
 *   Parameters:      EPNum: Endpoint Number
 *                    EPNum.0..3: Address
 *                    EPNum.7:    Dir
 *   Return Value:    None
 */
void uhw_DisableEP(u_char EPNum)
{
	u_long imf;

	imf = IntDisable();

	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(EPSTAT_DA));

	IntRestore(imf);

	return;
}

/*
 * Reset USB Endpoint
 *   Parameters:      EPNum: Endpoint Number
 *                    EPNum.0..3: Address
 *                    EPNum.7:    Dir
 *   Return Value:    None
 */
void uhw_ResetEP(u_char EPNum)
{
	u_long imf;

	imf = IntDisable();

	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(0));

	IntRestore(imf);

	return;
}

/*
 * Set Stall for USB Endpoint
 *   Parameters:      EPNum: Endpoint Number
 *                    EPNum.0..3: Address
 *                    EPNum.7:    Dir
 *   Return Value:    None
 */
void uhw_SetStallEP(u_char EPNum)
{
	u_long imf;

	imf = IntDisable();

	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(EPSTAT_STALL));

	IntRestore(imf);

	return;
}

/*
 * Clear Stall for USB Endpoint
 *   Parameters:      EPNum: Endpoint Number
 *                    EPNum.0..3: Address
 *                    EPNum.7:    Dir
 *   Return Value:    None
 */
void uhw_ClrStallEP(u_char EPNum)
{
	u_long imf;

	imf = IntDisable();

	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(0));

	IntRestore(imf);

	return;
}

/*
 * Write USB Endpoint Data
 *   Parameters:      EPNum: Endpoint Number
 *                    EPNum.0..3: Address
 *                    EPNum.7:    Dir
 *                    pData: Pointer to Data Buffer
 *                    cnt:   Number of bytes to write
 *   Return Value:    Number of bytes written
 */
void uhw_WriteEP(u_char EPNum, u_char *pData, u_long cnt)
{
	u_long imf;
	u_long n;

	imf = IntDisable();

	USB_CTRL = ((EPNum & 0x0F) << 2) | UCTRL_WR_EN;

	TX_PLENGTH = cnt;

	cnt = (cnt + 3) / 4;
	for(n=0; n<cnt; n++) {
		TX_DATA = (pData[0]<<0) + (pData[1]<<8) + (pData[2]<<16) + (pData[3]<<24);
		pData += 4;
	}

	USB_CTRL = 0;

	IntRestore(imf);

	return;
}

/*
 * validate OUT endpoint buffer
 */
void uhw_arm_WriteEP(u_char EPNum)
{
	u_long imf;

	imf = IntDisable();

	ucmd_no_data(CMD_SEL_EP(get_phy_EPaddr(EPNum)));
	ucmd_no_data(CMD_VALID_BUF);

	IntRestore(imf);

	return;
}

/*
 * Read USB Endpoint Data
 *   Parameters:      EPNum: Endpoint Number
 *                    EPNum.0..3: Address
 *                    EPNum.7:    Dir
 *                    pData: Pointer to Data Buffer
 *   Return Value:    Number of bytes read
 */
u_long uhw_ReadEP(u_char EPNum, u_char *pData, int buf_size)
{
	u_long imf;
	u_long cnt, n, rx;

	imf = IntDisable();

	USB_CTRL = ((EPNum & 0x0F) << 2) | UCTRL_RD_EN;

	do {
		cnt = RX_PLENGTH;
	} while((cnt & PKT_RDY) == 0);
	cnt &= PKT_LENGTH_MASK;

	for(n=0; n<cnt; n++) {
		if(--buf_size < 0)
			break;

		if((n % 4) == 0) {
			rx = RX_DATA;
		}
		*pData++ = rx & 0xff;
		rx >>= 8;
	}

	USB_CTRL = 0;

	IntRestore(imf);

	return(cnt);
}

/*
 * introduce hardwre buffer to receive next
 */
void uhw_ReadEP_finish(u_char EPNum)
{
	u_long imf;

	imf = IntDisable();

	ucmd_no_data(CMD_SEL_EP(get_phy_EPaddr(EPNum)));
	ucmd_no_data(CMD_CLR_BUF);

	IntRestore(imf);

	return;
}

/*
 * Write Command Data
 *   Parameters:      cmd:   Command
 *                    val:   Data
 *   Return Value:    None
 */
static void ucmd_write_data(u_long cmd, u_long val)
{
	DEV_INT_CLR = UDI_CCEMTY_INT;
	CMD_CODE = cmd;
	while((DEV_INT_STAT & UDI_CCEMTY_INT) == 0)
		;

	DEV_INT_CLR = UDI_CCEMTY_INT;
	CMD_CODE = val;
	while((DEV_INT_STAT & UDI_CCEMTY_INT) == 0)
		;
}

/*
 * Read Command Data
 *   Parameters:      cmd:   Command
 *   Return Value:    Data Value
 */
static u_long ucmd_read_data(u_long cmd)
{
	u_long val;

	DEV_INT_CLR = UDI_CCEMTY_INT | UDI_CDFULL_INT;
	CMD_CODE = cmd;
	while((DEV_INT_STAT & UDI_CCEMTY_INT) == 0)
		;
	while((DEV_INT_STAT & UDI_CDFULL_INT) == 0)
		;
	val = CMD_DATA;

	return(val);
}

/*
 * Write Command
 *   Parameters:      cmd:   Command
 *   Return Value:    None
 */
static void ucmd_no_data(u_long cmd)
{
	DEV_INT_CLR = UDI_CCEMTY_INT;
	CMD_CODE = cmd;
	while((DEV_INT_STAT & UDI_CCEMTY_INT) == 0)
		;
}

/*
 * Get Endpoint Address
 *   Parameters:      EPNum: Endpoint Number
 *                    EPNum.0..3: Address
 *                    EPNum.7:    Dir
 *   Return Value:    Endpoint Physical Address
 */
static u_long get_phy_EPaddr(u_long EPNum)
{
	u_long val;

	val = (EPNum & 0x0F) << 1;
	if(EPNum & 0x80) {
		val += 1;
	}
	return(val);
}

/* =============================================================== */

/*
 * USB Interrupt Service Routine
 */
static void USB_ISR(void)
{
	u_long dis, eis;
	u_long val, i, n;

FIO1SET = DEBUG_PORT2;		/* set high */

	dis = DEV_INT_STAT;				/* Device Interrupt Status */
RPRINTF("DEV_INT_STAT:%04x\r\n", dis);
	dis &= DEV_INT_EN;

	if(dis & UDI_DEV_STAT_INT) {
		/*
		 * Device Status Interrupt (Reset, Suspend/Resume, Connect change)
		 *   get device status detail
		 */
		ucmd_no_data(CMD_GET_DEV_STAT);
		val = ucmd_read_data(DAT_GET_DEV_STAT);
RPRINTF(" val:%04x\r\n", val);

		if(val & DSTAT_RST) {
			/*
			 * USB Reset
			 *   force clear all other interrupt requests
			 */
			EP_INT_CLR  = 0xFFFFFFFF;
			DEV_INT_CLR = 0xFFFFFFFF;

			/*
			 * reset params on other layer
			 */
			GotSUD = FALSE;
			GotCOD = FALSE;
			Sleep = FALSE;
			ucore_ResetCore();
			if(USB_Jump_Table[USB_JT_RESET] != NULL) {
				USB_Jump_Table[USB_JT_RESET](0);
			}
			goto isr_end;
		}

		if(val & DSTAT_SUSP_CH) {
			/*
			 * Suspend/Resume
			 *   these event is generated from 'USB_need_clock',
			 *   in other word, from USB bus activities.
			 */
			if(val & DSTAT_SUSP) {
				/*
				 * Suspend
				 *   interrupt is generated after confirmation
				 *   by 2sec-delay-filter.
				 */
				GotSUD = FALSE;
				GotCOD = FALSE;
				Sleep = TRUE;
				if(USB_Jump_Table[USB_JT_SUSPEND] != NULL) {
					USB_Jump_Table[USB_JT_SUSPEND](0);
				}
			} else {
				/*
				 * Resume
				 *   interrupt is immediately generated.
				 *   Resume interrupt will also come when power-on because
				 *   power-on default value of 'USB_need_clock' may be '1'.
				 *     (normaly with Reset interrupt, and ignored by this code)
				 */
//				GotSUD = FALSE;
//				GotCOD = FALSE;
//				Sleep = FALSE;
				if(USB_Jump_Table[USB_JT_RESUME] != NULL) {
					USB_Jump_Table[USB_JT_RESUME](0);
				}
			}

			goto isr_end;
		}

		if(val & DSTAT_CONN_CH) {
			/*
			 * Connect change (Vbus voltage appeared/disappeared)
			 */
//			GotSUD = FALSE;
//			GotCOD = FALSE;
//			Sleep = FALSE;
			if(USB_Jump_Table[USB_JT_POWER] != NULL) {
				USB_Jump_Table[USB_JT_POWER](val & DSTAT_CONN);
			}
			goto isr_end;
		}
	}

	if(dis & UDI_FRAME_INT) {
		/*
		 * Start of Frame Interrupt
		 */
		if(USB_Jump_Table[USB_JT_SOF] != NULL) {
			ucmd_no_data(CMD_RD_FRAME);
			val = ucmd_read_data(DAT_RD_FRAME);
			val = val | (ucmd_read_data(DAT_RD_FRAME) << 8);
			USB_Jump_Table[USB_JT_SOF](val);
		}
	}

	if(dis & UDI_ERR_INT) {
		/*
		 * Error Interrupt
		 */
		if(USB_Jump_Table[USB_JT_ERROR] != NULL) {
			ucmd_no_data(CMD_RD_ERR_STAT);
			val = ucmd_read_data(DAT_RD_ERR_STAT);

			if        (val & 0x01) {
				val = USB_ERR_PID;
			} else if(val & 0x02) {
				val = 0x100;	// Unexpected Error
			} else if(val & 0x04) {
				val = USB_ERR_CRC;
			} else if(val & 0x08) {
				val = USB_ERR_TIMEOUT;
			} else if(val & 0x10) {
				val = USB_ERR_EOP;
			} else if(val & 0x20) {
				val = 0x101;	// Buffer Overrun
			} else if(val & 0x40) {
				val = USB_ERR_BIT_STUFF;
			} else if(val & 0x80) {
				val = USB_ERR_DATA_TOGGLE;
			}
			USB_Jump_Table[USB_JT_ERROR](val);
		}
	}

	if(dis & UDI_EP_SLOW_INT) {
		/*
		 * Ednpoint's Slow Interrupt
		 */
		while((eis = EP_INT_STAT)) {
RPRINTF( "eis:%08x\r\n", eis);
			/*
			 * Check All Endpoints
			 */
	/*
	 * ABOUT HANDSHAKE WITH HOST BY SETUP REQUEST FLOW. (to debug with uart)
	 *
	 * the endpoint interrupt is generated when:
	 *   [SETUP], [OUT] ... when the data arrived
	 *   [IN]           ... after the data transmitted
	 *
	 * USB hardware may act like these.
	 *   [SETUP] ... always accepted, if previous [OUT] data is still
	 *               in hardware buffer, the data is overwritten.
	 *   [IN]    ... host should wait until the firmware set the data.
	 *   [OUT]   ... always accepted if the hardware buffer is free.
	 *
	 * SETUP request flows like these.
	 *   1.Control-Write:
	 *         [SETUP] [IN]..[IN(*)] [OUT as stat]  --> [nextSETUP]
	 *                             ^1            ^2               ^3
	 *   2.Control-Read:
	 *         [SETUP] [OUT]..[OUT] [IN as stat(*)] --> [nextSETUP]
	 *                                            ^4              ^5
	 *   3.Control-no-data:
	 *         [SETUP] [IN as stat(*)] --> [nextSETUP]
	 *                               ^6              ^7
	 * so, if the firmware can not follow the interrupt realtimely,
	 * multiple interrupt request may be detected.
	 *   case-A) if the firmware delay catching ^1, both IN+OUT request
	 *           appeared behind ^2.
	 *   case-B) more about case-A, if the firmware delay catching ^1+^2,
	 *           both IN+OUT request appeared behind ^3, with
	 *           'SETUP-overlapped' status.
	 *   case-C) if the firmware catch ^1 but delay catching ^2, OUT request
	 *           is overwritten behind ^3, with 'SETUP-overlapped' status.
	 *   case-D) if the firmware delay catching ^4, both IN+OUT request
	 *           appeared behind ^5.
	 *   case-E) if the firmware delay catching ^6, both IN+OUT request
	 *           appeared behind ^7.
	 *
	 * in those case, we shall proceed [IN] request first before [OUT] request.
	 * (we may not have jobs about this finish-of-last-IN)
	 *
	 * at the result, after armimg [IN(*)] endpoint, we can NOT make the
	 * host wait to proceed next.
	 *
	 * also, you shall care about synchronize polling context and interrupt
	 * handler context because I will process SETUP request in polling context.
	 */
		/*
		 * ܂ UART ŃfobObZ[WĂĂoĂŎԂ
		 * ĂŜ̓ȂȂ(ƁASETUP NGXg
		 * |[O[`ŏ邽)ɂ́A炪uɐiłv
		 * ƂȂzXg҂ĂKv̂łAꂪ
		 * ł̂͌ꂽƂɌ肳A킯łB
		 * IN ͂炪f[^pӂ܂ł̓zXg͑҂
		 * Ȃ킯łA(obt@̎)OUT endpoint 
		 * n[hM OK ႤƃzXg͎ɍsႤ킯ŁA
		 * u݃nhvɍsĂ݂҂鎟()
		 * Cxg܂ŋNĂvꍇPAKv킯łB
		 */
			for(i=0; i<LPC_N_EP; i++) {
				/*
				 * exchange priority for EP0in/EP0out
				 */
				if(i==0)
					n = 1;
				else if(i==1)
					n = 0;
				else
					n = i;

				if(eis & (1<<n)) {
//					DEV_INT_CLR = UDI_CDFULL_INT;
					EP_INT_CLR = 1 << n;
					while((DEV_INT_STAT & UDI_CDFULL_INT) == 0)
						;
					val = CMD_DATA;

					if(n == 0) {
						/* control out */
RPRINTF(" val:%02x\r\n", val);
						if(val & 0x04) {
							/* this is Setup Packet */
							GotSUD = TRUE;
							GotCOD = FALSE;
							USB_Jump_Table[USB_JT_SUDAV](0);
						} else {
							ucore_isr_EP0_out(0);
						}
					} else if(n == 1) {
						/* control in */
						ucore_isr_EP0_in(0);
					} else {
						/* generic endpoints */
						if(USB_Jump_Table_EP[n] != NULL) {
							USB_Jump_Table_EP[n](0);
						}
					}
				}
			}
		}
	}

isr_end:
	/*
	 * when jumping to here from 'detecting-Device-Status-change',
	 * ignore other interrupts if exist. (except UDI_EP_SLOW_INT)
	 */
	DEV_INT_CLR = dis;

	VICVectAddr = 0;				/* Acknowledge Interrupt */
FIO1CLR = DEBUG_PORT2;		/* set low */
}
